import os
from sidecar import Sidecar
from ipywidgets import Accordion, Select, VBox, Button, Output, Layout
from IPython.display import display, HTML


class Databases(object):
    """Singleton for the Databases object"""

    instance = None

    class __Databases(object):
        """Database browser implementation
            
        Args:
            spark (SparkSession): Spark Session object
        """
        def __init__(self, spark):
            self.spark = spark

        def create(self):
            """Create the sidecar view"""
            self.sc = Sidecar(title="Databases-%s" % os.environ["DBJL_CLUSTER"].split("-")[-1],
                              layout=Layout(width="300px"))
            self.refresh = Button(description="refresh")
            self.refresh.on_click(self.on_refresh)
            self.output = Output(layout=Layout(height="600px", width="320px", overflow_x="scroll", overflow_y="scroll"))
            self.output.add_class("db-detail")
            self.selects = []
            self.accordion = Accordion(children=[])

            with self.sc:
                display(VBox([self.refresh, self.accordion, self.output]))

            self.update()
            self.set_css()

        def on_refresh(self, b):
            """Refresh handler
            
            Args:
                b (ipywidgets.Button): clicked button
            """
            self.selects = []
            self.update()

        def update(self):
            """Update the view when an element was selected"""
            tables = {}
            for obj in self.spark.sql("show tables").rdd.collect():
                db = obj[0]
                table = obj[1]
                temp = obj[2]
                if temp and db == "":
                    db = "temp"
                if tables.get(db, None) is None:
                    tables[db] = []
                if temp:
                    tables[db].append("%s (temp)" % table)
                else:
                    tables[db].append(table)

            for db in sorted(tables.keys()):
                select = Select(options=[""] + sorted(tables[db]), disabled=False)
                select.observe(self.on_click(db, self), names='value')
                self.selects.append(select)
            self.accordion.children = self.selects
            for i, db in enumerate(sorted(tables.keys())):
                self.accordion.set_title(i, db)

        def on_click(self, db, parent):
            """Click handler providing db and parent as context
            
            Args:
                db (str): database name
                parent (object): parent object
            """
            def f(change):
                if change["old"] is not None:
                    parent.output.clear_output()
                    with parent.output:
                        if db == "temp":
                            table = change["new"]
                        else:
                            table = "%s.%s" % (db, change["new"])
                        if table.endswith(" (temp)"):
                            table = table[:-7]

                        try:
                            schema = parent.spark.sql("describe extended %s" % table)
                            rows = int(parent.spark.conf.get("spark.sql.repl.eagerEval.maxNumRows"))
                            parent.spark.conf.set("spark.sql.repl.eagerEval.maxNumRows", 1000)
                            display(schema)
                            parent.spark.conf.set("spark.sql.repl.eagerEval.maxNumRows", rows)
                        except:
                            print("schema cannot be accessed, table most probably broken")

            return f

        def close(self):
            """Close view"""
            self.selects = []
            self.sc.close()

        def set_css(self):
            """Set CSS"""
            display(
                HTML("""
            <style>
            .db-detail .p-Widget {
            overflow: visible;
            }
            </style>
            """))

    def __init__(self, spark=None):
        """Singleton initializer
        
        Args:
            spark (SparkSession, optional): Spark Session object. Defaults to None.
        """
        if not Databases.instance:
            Databases.instance = Databases.__Databases(spark)

    def __getattr__(self, name):
        """Singleton getattr overload"""
        return getattr(self.instance, name)



import atexit
import base64
import configparser
import getpass
import io
import os
import random
import sys
import time

from IPython.core.magic import line_magic, line_cell_magic
from IPython import get_ipython
from IPython.display import HTML, display

sys.path.insert(0, "/databricks/spark/python/lib/py4j-0.10.7-src.zip")
sys.path.insert(0, "/databricks/spark/python")
sys.path.insert(0, "/databricks/jars/spark--driver--spark--resources-resources.jar")

# will only work on the Databricks side
try:
    from pyspark.context import SparkContext
    from pyspark.conf import SparkConf
    from pyspark.sql import HiveContext

    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Suppress py4j loading message on stderr by redirecting sys.stderr
        stderr_orig = sys.stderr
        sys.stderr = io.StringIO()
        from PythonShell import get_existing_gateway, RemoteContext  # pylint: disable=import-error
        out = sys.stderr.getvalue()
        # Restore sys.stderr
        sys.stderr = stderr_orig
        # Print any other error message to stderr
        if not "py4j imported" in out:
            print(out, file=sys.stderr)

    from dbutils import DBUtils  # pylint: disable=import-error
except:
    pass

from databrickslabs_jupyterlab.rest import Command, DatabricksApiException
from databrickslabs_jupyterlab.progress import load_progressbar, load_css
from databrickslabs_jupyterlab.dbfs import Dbfs
from databrickslabs_jupyterlab.database import Databases


class JobInfo():
    """Job info class for Spark jobs
    Args:
        pool_id (str): Pool ID to separate Spark jobs from each other
        group_id (str): Group ID to enable killing jobs and progress bar
    """
    def __init__(self, pool_id):
        self.pool_id = pool_id
        self.group_id = None


class DatabricksBrowser:
    """[summary]
    Args:
        spark (SparkSession): Spark Session object
        dbutils (DBUtils): DbUtils object
    """
    def __init__(self, spark, dbutils):
        self.spark = spark
        self.dbutils = dbutils

    def dbfs(self):
        """Start dbfs browser"""
        Dbfs(self.dbutils).create()

    def databases(self):
        """Start Database browser"""
        Databases(self.spark).create()


def is_remote():
    """Check whether the current context is on the remote cluster
    
    Returns:
        bool: True if remote else False
    """
    return os.environ.get("DBJL_HOST", None) is not None


def is_azure():
    """Check whether the current context is on Azure Databricks
    
    Returns:
        bool: True if Azure Databricks else False
    """
    return os.environ.get("DBJL_ORG", None) is not None


def dbcontext(progressbar=True):
    """Create a databricks context
    The following objects will be created
    - Spark Session
    - Spark Context
    - Spark Hive Context
    - DBUtils (fs module only)
    
    Args:
        progressbar (bool, optional): If True the spark progressbar will be installed. Defaults to True.
    """
    def get_sparkui_url(host, organisation, clusterId):
        if organisation is None:
            sparkUi = "%s#/setting/clusters/%s/sparkUi" % (host, clusterId)
        else:
            sparkUi = "%s/?o=%s#/setting/clusters/%s/sparkUi" % (host, organisation, clusterId)
        return sparkUi

    def show_status(spark, sparkUi):
        output = """
        <div>
            <dl>
            <dt>Spark Version</dt><dd>{sc.version}</dd>
            <dt>Spark Application</dt><dd>{sc.appName}</dd>
            <dt>Spark UI</dt><dd><a href="{sparkUi}">go to ...</a></dd>
            </dl>
        </div>
        """.format(sc=spark.sparkContext,
                   sparkUi=get_sparkui_url(host, organisation, clusterId),
                   num_executors=len(spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos()))
        display(HTML(output))

    # Get the configuration injected by the client
    #
    profile = os.environ.get("DBJL_PROFILE", None)
    host = os.environ.get("DBJL_HOST", None)
    clusterId = os.environ.get("DBJL_CLUSTER", None)
    organisation = os.environ.get("DBJL_ORG", None)

    sparkUi = get_sparkui_url(host, organisation, clusterId)

    if not is_remote():
        return "This is not a remote Databricks kernel"

    ip = get_ipython()
    spark = ip.user_ns.get("spark")
    if spark is not None:
        print("Spark context already exists")
        load_css()
        show_status(spark, sparkUi)
        return None

    # Create a Databricks virtual python environment and start thew py4j gateway
    #
    token = getpass.getpass("Creating a Spark execution context:\nEnter personal access token for profile '%s'" % profile)

    try:
        command = Command(url=host, cluster_id=clusterId, token=token)
    except DatabricksApiException as ex:
        print(ex)
        return None

    print("Gateway created for cluster '%s' " % (clusterId), end="", flush=True)

    # Fetch auth_token and gateway port ...
    #
    cmd = 'c=sc._gateway.client.gateway_client; print(c.gateway_parameters.auth_token + "|" + str(c.port))'
    result = command.execute(cmd)
    
    if result[0] != 0:
        print(result[1])
        return None
    
    auth_token, port = result[1].split("|")
    port = int(port)

    interpreter = "/databricks/python/bin/python"
    # Ensure that driver and executors use the same python
    #
    os.environ["PYSPARK_PYTHON"] = interpreter
    os.environ["PYSPARK_DRIVER_PYTHON"] = interpreter

    # ... and connect to this gateway
    #
    gateway = get_existing_gateway(port, True, auth_token)
    print(". connected")
    # print("Python interpreter: %s" % interpreter)

    # Retrieve spark session, sqlContext and sparkContext
    #
    conf = SparkConf(_jconf=gateway.entry_point.getSparkConf())
    sqlContext = RemoteContext(gateway=gateway, conf=conf)

    sqlContext = HiveContext(sqlContext, gateway.entry_point.getSQLContext())
    spark = sqlContext.sparkSession
    sc = spark.sparkContext

    # Enable pretty printing of dataframes
    #
    spark.conf.set("spark.sql.repl.eagerEval.enabled", "true")

    # Define a separate pool for the fair scheduler
    # Todo: Find a better way to store pool_id instead of this hack
    #
    job_info = JobInfo(str(random.getrandbits(64)))

    # Patch the remote spark UI into the _repr_html_ call
    #
    def repr_html(uiWebUrl):
        def sc_repr_html():
            return """
            <div>
                <p><b>SparkContext</b></p>
                <p><a href="{uiWebUrl}">Spark UI</a></p>
                <dl>
                  <dt>Version</dt><dd><code>v{sc.version}</code></dd>
                  <dt>AppName</dt><dd><code>{sc.appName}</code></dd>
                  <dt>Master</dt><dd><code>{sc.master}</code></dd>
                </dl>
            </div>
            """.format(sc=spark.sparkContext, uiWebUrl=uiWebUrl)

        return sc_repr_html

    sc_repr_html = repr_html(sparkUi)
    sc._repr_html_ = sc_repr_html

    # Monkey patch Databricks Cli to allow mlflow tracking with the credentials provided
    # by this routine
    # Only necessary when mlflow is installed
    #
    try:
        from databricks_cli.configure.provider import ProfileConfigProvider, DatabricksConfig

        def get_config(self):
            config = DatabricksConfig(host, None, None, token, False)
            if config.is_valid:
                return config
            return None

        ProfileConfigProvider.get_config = get_config
    except:
        pass

    # Initialize the ipython shell with spark context
    #
    shell = get_ipython()
    shell.sc = sc
    shell.sqlContext = sqlContext
    shell.displayHTML = lambda html: display(HTML(html))

    # Retrieve the py4j gateway entrypoint
    #
    entry_point = spark.sparkContext._gateway.entry_point

    # Initialize dbutils
    #
    dbutils = DBUtils(shell, entry_point)

    # Setting up Spark progress bar
    #
    if progressbar:
        # print("Set up Spark progress bar")
        load_progressbar(ip, sc, job_info)
        load_css()

    # Register sql magic
    #
    ip.register_magic_function(sql, magic_kind='line_cell')

    # Ensure that the virtual python environment and py4j gateway gets shut down
    # when the python interpreter shuts down
    #
    def shutdown_kernel(command):
        def handler():
            from IPython import get_ipython
            ip = get_ipython()

            ip = get_ipython()
            if ip.user_ns.get("spark", None) is not None:
                del ip.user_ns["spark"]
            if ip.user_ns.get("sc", None) is not None:
                del ip.user_ns["sc"]
            if ip.user_ns.get("sqlContext", None) is not None:
                del ip.user_ns["sqlContext"]
            if ip.user_ns.get("dbutils", None) is not None:
                del ip.user_ns["dbutils"]
            # Context is a singleton
            command.close()

        return handler


    atexit.register(shutdown_kernel(command))

    # Forward spark variables to the user namespace
    #
    ip.user_ns["spark"] = spark
    ip.user_ns["sc"] = sc
    ip.user_ns["sqlContext"] = sqlContext
    ip.user_ns["dbutils"] = dbutils
    ip.user_ns["dbbrowser"] = DatabricksBrowser(spark, dbutils)

    print("The following global variables have been created:")
    print("- spark       Spark session")
    print("- sc          Spark context")
    print("- sqlContext  Hive Context")
    print("- dbutils     Databricks utilities (filesystem access only)")
    print("- dbbrowser   Allows to browse dbfs and databases:")
    print("              - dbbrowser.dbfs()")
    print("              - dbbrowser.databases()\n")

    show_status(spark, sparkUi)
    return None


@line_cell_magic
def sql(line, cell=None):
    """Cell magic th execute SQL commands
    
    Args:
        line (str): line behind %sql
        cell (str, optional): cell below %sql. Defaults to None.
    
    Returns:
        DataFrame: DataFrame of the SQL result
    """
    ip = get_ipython()
    spark = ip.user_ns["spark"]
    if cell == None:
        code = line
    else:
        code = cell
    return spark.sql(code)



import mlflow
import mlflow.sklearn
from mlflow.tracking import MlflowClient
from IPython.display import HTML, display
import os
import pandas as pd
import datetime
import tempfile
import warnings
from functools import reduce
from databrickslabs_jupyterlab.connect import is_remote


class GridSearchCV():
    """Abstracts away the differences between spark_sklearn and scikit-learn GridSearch
        
    Args:
        estimator (Scikit Learn Estimator): Estimator used in the model
        param_grid (dict): Dictionary of the paramter grid for grid search
        spark (SparkSession, optional): SparkSession object. Defaults to None.
    """
    def __init__(self, estimator, param_grid, *args, spark=None, **kwargs):
        self.estimator = estimator
        self.grid_size = reduce(lambda a, b: a * b, [len(p) for p in param_grid.values()])
        self.results = None

        if is_remote():
            import spark_sklearn  # pylint: disable=import-error
            self.gs = spark_sklearn.GridSearchCV(spark.sparkContext, estimator, param_grid, *args, **kwargs)
        else:
            import sklearn.model_selection
            self.gs = sklearn.model_selection.GridSearchCV(estimator, param_grid, *args, **kwargs)

    def fit(self, x, y):
        """Fitting routine for grid search
        
        Args:
            x (list, pandas Dataframe): The features x
            y (list, pandas Dataframe): The target y
        
        Returns:
            Scikit-learn model: The model created by fitting the data
        """
        if is_remote():
            print("Remote crossvalidation,", end=" ")
        else:
            print("Local crossvalidation,", end=" ")
        print("paramter grid size: %d\n" % self.grid_size)

        self.results = self.gs.fit(x, y)
        return self.results

    def log_cv(self, experiment, name, tracking_uri=None):
        """Logging of cross validation results to mlflow tracking server
        
        Args:
            experiment (str): experiment ID
            name (str): Name of the experiment artifact (prefix)
            tracking_uri (str, optional): URI of the tracking server. 
                                         Defaults to None, which will use remote tracking in remote case
        """
        cv_results = self.results.cv_results_
        best = self.results.best_index_

        timestamp = datetime.datetime.now().isoformat().split(".")[0].replace(":", ".")

        num_runs = len(cv_results["rank_test_score"])
        run_name = "run %d (best run of %d):" % (self.results.best_index_, num_runs)

        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)

        mlflow.set_experiment(experiment)

        with mlflow.start_run(run_name=run_name):  #  as run:

            mlflow.log_param("folds", self.results.cv)

            print("Logging parameters")
            params = list(self.results.param_grid.keys())
            for param in params:
                mlflow.log_param(param, cv_results["param_%s" % param][best])

            print("Logging metrics")
            mlflow.log_metric("mean_test_score", cv_results["mean_test_score"][best])
            mlflow.log_metric("std_test_score", cv_results["std_test_score"][best])

            print("Logging model")
            mlflow.sklearn.log_model(self.results.best_estimator_, "model")

            print("Logging CV results matrix")
            tempdir = tempfile.TemporaryDirectory().name
            os.mkdir(tempdir)
            filename = "%s-%s-cv_results.csv" % (name, timestamp)
            csv = os.path.join(tempdir, filename)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                pd.DataFrame(cv_results).sort_values(by='rank_test_score').to_csv(csv, index=False)

            mlflow.log_artifact(csv, "cv_results")

        client = MlflowClient()
        experiment_id = client.get_experiment_by_name(experiment).experiment_id

        if is_remote():
            if os.environ.get("DBJL_ORG", None) is None:
                display(
                    HTML('<a href=%s/#mlflow/experiments/%s>Goto experiment</a>' %
                         (os.environ["DBJL_HOST"], experiment_id)))
            else:
                display(
                    HTML('<a href=%s?o=%s#mlflow/experiments/%s>Goto experiment</a>' %
                         (os.environ["DBJL_HOST"], os.environ["DBJL_ORG"], experiment_id)))
        else:
            display(HTML('<a href=%s/#/experiments/%s>Goto experiment</a>' % (tracking_uri, experiment_id)))



import base64
import json
import os
import re
import sys
import time
from os.path import expanduser
import socket
import subprocess
import uuid
import glob

import inquirer
import requests
from urllib.parse import unquote

from databricks_cli.configure.provider import get_config, ProfileConfigProvider
from databricks_cli.sdk.api_client import ApiClient
from databricks_cli.clusters.api import ClusterApi
from databricks_cli.libraries.api import LibrariesApi

import databrickslabs_jupyterlab
from databrickslabs_jupyterlab._version import __version__
from databrickslabs_jupyterlab.rest import Command, DatabricksApiException
from databrickslabs_jupyterlab.utils import bye, Dark, print_ok, print_error, print_warning
from databrickslabs_jupyterlab.local import execute, get_local_libs, utf8_decode


def connect(profile):
    """Initialize Databricks API client
    
    Args:
        profile (str): Databricks CLI profile string
    
    Returns:
        ApiClient: Databricks ApiClient object
    """
    config = ProfileConfigProvider(profile).get_config()
    if config is None:
        print_error("Cannot initialize ApiClient")
        bye(1)
    verify = config.insecure is None
    if config.is_valid_with_token:
        return ApiClient(host=config.host, token=config.token, verify=verify)
    else:
        print_error("Token for profile '%s' is invalid" % profile)
        bye(1)


def select_cluster(clusters):
    """Build a list of clusters for the user to select one
    
    Args:
        clusters (list): list of clusters
    
    Returns:
        dict: Cluster config of selected cluster
    """
    def entry(i, cluster):
        if cluster.get("autoscale", None) is None:
            return "%s: '%s' (id: %s, state: %s, workers: %d)" % (i, cluster["cluster_name"], cluster["cluster_id"],
                                                                cluster["state"], cluster["num_workers"])
        else:
            return "%s: '%s' (id: %s, state: %s, scale: %d-%d)" % (i, cluster["cluster_name"], cluster["cluster_id"],
                                                                 cluster["state"], cluster["autoscale"]["min_workers"],
                                                                 cluster["autoscale"]["max_workers"])

    choice = [
        inquirer.List('cluster_id',
                      message='Which cluster to connect to?',
                      choices=[entry(i, cluster) for i, cluster in enumerate(clusters)])
    ]
    answer = inquirer.prompt(choice, theme=Dark())
    return clusters[int(answer["cluster_id"].split(":")[0])]


def get_cluster(profile, host, token, cluster_id=None, status=None):
    """Get the cluster configuration from remote
    
    Args:
        profile (str): Databricks CLI profile string
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile stringf
        cluster_id (str, optional): If cluster_id is given, the user will not be asked to select one. Defaults to None.
        status (Status, optional): A Status class providing set_status. Defaults to None.
    
    Returns:
        tuple: Cluster configs: cluster_id, public_ip, cluster_name and a flag whether it was started
    
    Returns:
        tuple: (cluster_id, public_ip, cluster_name, started)
    """
    with open("%s/.ssh/id_%s.pub" % (expanduser("~"), profile)) as fd:
        try:
            ssh_pub = fd.read().strip()
        except:
            print_error("   Error: ssh key for profile 'id_%s.pub' does not exist in %s/.ssh" % (profile, expanduser("~")))
            bye()

    try:
        apiclient = connect(profile)
        client = ClusterApi(apiclient)
        clusters = client.list_clusters()
    except Exception as ex:
        print_error(ex)
        return (None, None, None, None)

    clusters = clusters["clusters"]

    if cluster_id is not None:
        cluster = None
        for c in clusters:
            if c["cluster_id"] == cluster_id:
                cluster = c
                break

        if cluster is None:
            print_error("   Error: A cluster with id '%s' does not exist in the workspace of profile '%s'" %
                        (cluster_id, profile))
            return (None, None, None, None)

        if ssh_pub not in [c.strip() for c in cluster.get("ssh_public_keys", [])]:
            print_error("   Error: Cluster with id '%s' does not have ssh key '~/.ssh/id_%s' configured" %
                        (cluster_id, profile))
            return (None, None, None, None)
    else:
        my_clusters = [
            cluster for cluster in clusters
            if ssh_pub in [c.strip() for c in cluster.get("ssh_public_keys", [])]
        ]

        if not my_clusters:
            print_error(
                "    Error: There is no cluster in the workspace for profile '%s' configured with ssh key '~/.ssh/id_%s':" %
                (profile, profile))
            print("    Use 'databrickslabs_jupyterlab %s -s' to configure ssh for clusters in this workspace\n" % profile)
            return (None, None, None, None)

        current_conda_env = os.environ.get("CONDA_DEFAULT_ENV", None)
        found = None
        for i, c in enumerate(my_clusters):
            if c["cluster_name"].replace(" ", "_") == current_conda_env:
                found = c["cluster_name"]
                break

        if found is not None:
            print_warning(
                "\n   => The current conda environment is '%s'.\n      You might want to select cluster %d with the name '%s'?\n" %
                (current_conda_env, i, found))

        cluster = select_cluster(my_clusters)

    cluster_id = cluster["cluster_id"]
    cluster_name = cluster["cluster_name"]

    try:
        response = client.get_cluster(cluster_id)
    except Exception as ex:
        print_error(ex)
        return (None, None, None, None)

    state = response["state"]
    if not state in ["RUNNING", "RESIZING"]:
        if state == "TERMINATED":
            print("   => Starting cluster %s" % cluster_id)
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting")
            try:
                response = client.start_cluster(cluster_id)
            except Exception as ex:
                print_error(ex)
                return (None, None, None, None)

        print("   => Waiting for cluster %s being started (this can take up to 5 min)" % cluster_id)
        print("   ", end="", flush=True)

        while not state in ("RUNNING", "RESIZING"):
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting", False)
            print(".", end="", flush=True)
            time.sleep(5)
            try:
                response = client.get_cluster(cluster_id)
            except Exception as ex:
                print_error(ex)
                return (None, None, None, None)

            if response.get("error", None) is not None:
                print_error(response["error"])
                return (None, None, None, None)
            else:
                state = response["state"]
        print_ok("\n   => OK")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster started")

        print("\n   => Waiting for libraries on cluster %s being installed (this can take some time)" % cluster_id)
        print("   ", end="", flush=True)

        done = False
        while not done:
            try:
                states = get_library_state(profile, cluster_id)
            except DatabricksApiException as ex:
                print_error(ex)
                return (None, None, None, None)
            installing = any([s in ["PENDING", "RESOLVING", "INSTALLING"] for s in states])
            if installing:
                if status is not None:
                    status.set_status(profile, cluster_id, "Installing cluster libraries", False)
                print(".", end="", flush=True)
                time.sleep(5)
            else:
                done = True
                print_ok("\n   => OK\n")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster libraries installed", False)

    public_ip = response["driver"].get("public_dns", None)
    if public_ip is None:
        print_error("   Error: Cluster does not have public DNS name")
        return (None, None, None, None)

    print_ok("   => Selected cluster: %s (%s)" % (cluster_name, public_ip))

    return (cluster_id, public_ip, cluster_name, None)


def get_python_path(host, conda_env=None):
    # conda_default_env = ssh(host, "echo $DEFAULT_DATABRICKS_ROOT_CONDA_ENV").strip().decode("utf-8")
    # if conda_default_env == "":
    #     python_path = "/databricks/python3/bin"
    #     print_ok("   => %s (pip managed cluster)" % python_path)
    # else:
    #     if conda_env is None:  # if there is no conda environment variable, use default
    #         conda_env = conda_default_env
    #     python_path = "/databricks/conda/envs/%s/bin" % conda_env
    #     print_ok("   => %s (conda managed cluster)" % python_path)

    python_path = "/databricks/python/bin"
    return python_path


def install_libs(cluster_id, host, token):
    """Install ipywidgets, sidecar and databrickslabs_jupyterlab libraries on the driver
    
    Args:
        host (str): host from databricks cli config for given profile string
        python_path (str): Remote python path to be used for kernel
    """
    python_path = get_python_path(cluster_id)

    packages = get_local_libs()
    deps = {p["name"]: p["version"] for p in packages if p["name"] in ["ipywidgets", "sidecar"]}
    libs = ["ipywidgets==%s" % deps["ipywidgets"],
            "sidecar==%s" % deps["sidecar"],
            "databrickslabs-jupyterlab==%s" % __version__]
    pip_cmd = ["%s/pip" %python_path,  "install", "-q", "--no-warn-conflicts", "--disable-pip-version-check"] + libs

    cmd =  'import subprocess, json; ' + \
          ('ret = subprocess.run(%s, stderr=subprocess.PIPE); ' % pip_cmd) + \
           'print(json.dumps([ret.returncode, str(ret.stderr)]))'

    print("   => Installing %s" % ", ".join(libs))
    print("   ", end="")
    try:
        command = Command(url=host, cluster_id=cluster_id, token=token)
        result = command.execute(cmd)
        command.close()
        print()
    except DatabricksApiException as ex:
        print("REST API error", ex)
        return False

    if result[0] == 0:
        return tuple(json.loads(utf8_decode(result[1])))
    else:
        return result


def get_remote_packages(cluster_id, host, token):
    cmd = 'import sys, pkg_resources, json; ' + \
          'print(json.dumps([{"name": "python", "version": "%d.%d" % (sys.version_info.major, sys.version_info.minor)}] + ' + \
          '[{"name":p.key, "version":p.version} for p in pkg_resources.working_set]))'
    try:
        command = Command(url=host, cluster_id=cluster_id, token=token)
        result = command.execute(cmd)
        command.close()
    except DatabricksApiException as ex:
        print(ex)
        return None
    return result

def is_reachable(public_dns):
    """Check whether a remote cluster is reachable
    
    Args:
        public_dns (str): Public IP address or DNS name
    
    Returns:
        bool: True if reachable else False
    """
    return subprocess.call(["nc", "-z", public_dns, "2200"]) == 0
    # sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # sock.settimeout(3)
    # result = sock.connect_ex((public_dns, 2200))
    # try:
    #     sock.close()
    # except:
    #     pass
    # return result == 0


def get_library_state(profile, cluster_id):
    """Get the state of the library installation on the remote cluster
    
    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile stringf
    
    Returns:
        list: list of installation status for each custom library
    """
    try:
        apiclient = connect(profile)
        client = LibrariesApi(apiclient)
        libraries = client.cluster_status(cluster_id)
    except Exception as ex:
        print_error(ex)
        return None

    if libraries.get("library_statuses", None) is None:
        return []
    else:
        return [lib["status"] for lib in libraries["library_statuses"]]


def check_installed(cluster_id, host, token):
    """Check whether databrickslabs_jupyterlab is installed on the remote host
    
    Args:
    
    Returns:
        bool: True if installed else False
    """
    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        packages = json.loads(result[1])
        package_names = [p["name"] for p in packages]
        needed = ["databrickslabs-jupyterlab", "ipywidgets", "ipykernel"]

        return all([p in package_names for p in needed])
    else:
        print(result[1])
        return False


def version_check(cluster_id, host, token, flag):
    """Compare local and remote library versions
    
    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        flag (str): all|diff|same
    """
    def normalize(key):
        return key.lower().replace("-", "_")

    packages = get_local_libs()
    deps = {normalize(p["name"]): p["version"] for p in packages}

    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        remote_packages = json.loads(result[1])
    else:
        return
    remote_deps = {normalize(p["name"]): p["version"] for p in remote_packages}
    joint_keys = sorted(list(set(list(deps.keys()) + list(remote_deps.keys()))))
    print("%-30s %-10s%-10s" % ("Package", "local", "remote"))
    if str(flag) == "all":
        scope = joint_keys
    elif str(flag) == "same":
        scope = [key for key in joint_keys if deps.get(key, None) == remote_deps.get(key, None)]
    else:
        scope = [
            key for key in joint_keys if deps.get(key, None) != remote_deps.get(key, None)
            #            and deps.get(key, None) is not None and remote_deps.get(key, None) is not None
        ]
    for key in scope:
        result = "%-30s %-10s  %-10s" % (key, deps.get(key, "--"), remote_deps.get(key, "--"))
        if deps.get(key) == remote_deps.get(key):
            print_ok(result)
        else:
            print_error(result)

def configure_ssh(profile, host, token, cluster_id):
    """Configure SSH for the remote cluster
    
    Args:
        profile (str): Databricks CLI profile string
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        cluster_id (str): cluster ID
    """
    sshkey_file = os.path.expanduser("~/.ssh/id_%s" % profile)
    if not os.path.exists(sshkey_file):
        print("\n   => ssh key '%s' does not exist" % sshkey_file)
        answer = input("   => Shall it be created (y/n)? (default = n): ")
        if answer.lower() == "y":
            print("   => Creating ssh key %s" % sshkey_file)
            result = execute(["ssh-keygen", "-b", "2048",  "-N", "", "-f",  sshkey_file])
            if result["returncode"] == 0:
                print_ok("   => OK")
            else:
                print_error(result["stderr"])
                bye()
        else:
            bye()
    else:
        print_ok("\n   => ssh key '%s' already exists" % sshkey_file)

    with open(sshkey_file + ".pub", "r") as fd:
        sshkey = fd.read().strip()

    try:
        apiclient = connect(profile)
        client = ClusterApi(apiclient)
    except Exception as ex:
        print_error(ex)
        return None

    try:
        response = client.get_cluster(cluster_id)
    except Exception as ex:
        print_error(ex)
        return None

    ssh_public_keys = response.get("ssh_public_keys", [])
    if sshkey in [key.strip() for key in ssh_public_keys]:
        print_ok("   => public ssh key already configured for cluster %s" % cluster_id)
        bye()

    request = {}
    for key in [
            "autotermination_minutes", "cluster_id", "cluster_name", "cluster_source", "creator_user_name",
            "default_tags", "driver_node_type_id", "enable_elastic_disk", "init_scripts_safe_mode", "node_type_id",
            "spark_env_vars", "spark_version"
    ]:
        request[key] = response[key]

    if response.get("aws_attributes", None) is not None:
        request["aws_attributes"] = response["aws_attributes"]

    if response.get("num_workers", None) is not None:
        request["num_workers"] = response["num_workers"]

    if response.get("autoscale", None) is not None:
        request["autoscale"] = response["autoscale"]

    request["ssh_public_keys"] = ssh_public_keys + [sshkey]

    print_warning("   => The ssh key will be added to the cluster. \n   Note: The cluster will be restarted immediately!")
    answer = input("   => Shall the ssh key be added and the cluster be restarted (y/n)? (default = n): ")
    if answer.lower() == "y":
        try:
            response = client.edit_cluster(request)
        except DatabricksApiException as ex:
            print_error(str(ex))
            return None
        print_ok("   => OK")
    else:
        print_error("   => Cancelled")

def download_notebook(url):
    """Download Databricks demo notebooks from docs.databricks.com.
    It adds two cells on the top to initialize the Databricks context and MLflow
    
    Args:
        url (str): The HTML url copied from the "Get notebook link"
    """

    def wrap(typ, cell):
        return {"cell_type":typ, "source": cell, "output":[], "metadata":{}, "execution_count":0 }

    cell1 = """# Retrieve Spark Context
from databrickslabs_jupyterlab.connect import dbcontext, is_remote
dbcontext()"""

    cell2 = """# Setup MLflow (optional)
import os
from mlflow
from mlflow.tracking import MlflowClient

def remote_mlflow_setup(experiment_name):
    mlflow.set_experiment(experiment_name)
    experiment = MlflowClient().get_experiment_by_name(experiment_name)    
    from IPython.display import HTML, display
    url = "%s/#mlflow/experiments/%s" % (os.environ["DBJL_HOST"], experiment.experiment_id)
    display(HTML('View the <a href="%s">MLflow experiment</a>' % url))    


experiment_name = <FILL>
home = "/Users/%s" % <FILL>
experiment_name = "%s/experiments/%s" % (home, experiment_name)
remote_mlflow_setup(experiment_name)"""

    print("*** experimental ***")
    name = os.path.splitext(os.path.basename(url))[0]

    resp = requests.get(url)
    if resp.status_code != 200:
        print("error")
    else:
        r = re.compile("<script>(.+?)</script>")
        scripts = r.findall(resp.text)
        script = [s for s in scripts if "__DATABRICKS_NOTEBOOK_MODEL" in s][0]
        r = re.compile("'(.+?)'")
        s = r.findall(script)
        nb = json.loads(unquote(base64.b64decode(s[0]).decode("utf-8")))
        cells = [c["command"] for c in nb["commands"]]
        ipy = {"cells":[wrap("code", cell1), wrap("code", cell2)],  "nbformat": 4, "nbformat_minor": 4, "metadata":{}}
        for c in cells:
            t = "code"            
            if c.startswith("%md"):
                t = "markdown"
                c = c[3:]
            elif c.startswith("%s"):
                # Jupyter uses a cell magic
                c = "%" + c
            ipy["cells"].append(wrap(t, c))
        with open("%s.ipynb" % name, "w") as fd:
            fd.write(json.dumps(ipy))
        print("Downloaded notebook to %s.ipynb" % name)



import os
from collections import defaultdict
from pexpect import pxssh
import random
import ssh_config
import sys
import time

from inquirer.themes import Default, term


def bye(status=0):
    """Standard exit function
    
    Args:
        msg (str, optional): Exit message to be printed. Defaults to None.
    """
    if status != 0:
        print_error("\n=> Exiting")
    sys.exit(status)


class Dark(Default):
    """Dark Theme for inquirer"""
    def __init__(self):
        super().__init__()
        self.List.selection_color = term.cyan


class Colors:
    ERROR = '\033[91m'
    OK = '\033[92m'
    WARNING = '\033[93m'
    RESET = '\033[0m'


def _print(color, *args):
    print(color, end="")
    print(*args, end="")
    print(Colors.RESET)


def print_ok(*args):
    _print(Colors.OK, *args)


def print_error(*args):
    _print(Colors.ERROR, *args)


def print_warning(*args):
    _print(Colors.WARNING, *args)


class SshConfig:
    def __init__(self):
        self.config_file = os.path.expanduser("~/.ssh/config")
        self.load()

    def _get_mtime(self):
        return os.stat(self.config_file).st_mtime

    def load(self):
        self.mtime = self._get_mtime()
        sc = ssh_config.SSHConfig.load(self.config_file)
        self.hosts = {
            h.name: h.attributes()["HostName"]
            for h in sc.hosts() if h.attributes().get("HostName", None) is not None
        }

    def get_dns(self, host_name):
        if self._get_mtime() > self.mtime:
            self.load()
        return self.hosts.get(host_name, None)





import os
import re

from sidecar import Sidecar
from ipywidgets import Select, VBox, HBox, Button, Output
from IPython.display import display


class Dbfs(object):
    """Singleton for the DBFS object"""

    instance = None

    class __Dbfs(object):
        """Database browser implementation
            
        Args:
            dbutils (DBUtils): DBUtils object (for fs only)
        """
        def __init__(self, dbutils):
            self.dbutils = dbutils

        def create(self):
            """Create the sidecar view"""
            self.sc = Sidecar(title="DBFS-%s" % os.environ["DBJL_CLUSTER"].split("-")[-1])
            self.path = "/"
            self.flist = Select(options=[], rows=40, disabled=False)
            self.flist.observe(self.on_click, names='value')

            self.refresh = Button(description="refresh")
            self.refresh.on_click(self.on_refresh)
            self.output = Output()

            self.up = Button(description="up")
            self.up.on_click(self.on_up)

            with self.sc:
                display(VBox([HBox([self.up, self.refresh]), self.flist, self.output]))

            self.update()

        def convertBytes(self, fsize):
            """Convert bytes to largest unit
            
            Args:
                fsize (int): Size in bytes
            
            Returns:
                tuple: size of largest unit, largest unit
            """
            size = fsize
            unit = "B"
            if size > 1024 * 1024 * 1024 * 10:
                size = int(size / 1024.0 / 1024.0 / 1024.0)
                unit = "GB"
            elif size > 1024 * 1024 * 10:
                size = int(size / 1024.0 / 1024.0)
                unit = "MB"
            elif size > 1024 * 10:
                size = int(size / 1024.0)
                unit = "KB"
            return (size, unit)

        def update(self):
            """Update the view when an element was selected"""
            with self.output:
                print("updating ...")
            fobjs = self.dbutils.fs.ls(self.path)
            self.show_path(self.path)

            dirs = sorted([fobj.name for fobj in fobjs if fobj.isDir()], key=lambda x: x.lower())
            files = sorted(
                ["%s (%d %s)" % ((fobj.name, ) + self.convertBytes(fobj.size)) for fobj in fobjs if not fobj.isDir()],
                key=lambda x: x[0].lower())
            self.flist.options = [""] + dirs + files

        def show_path(self, path):
            """Show path in output widget
            
            Args:
                path (str): Currently selected path
            """
            self.output.clear_output()
            with self.output:
                print("dbfs:" + re.sub(r"\s\(.*?\)$", "", path))

        def on_refresh(self, b):
            """Refresh handler
            
            Args:
                b (ipywidgets.Button): clicked button
            """
            self.update()

        def on_up(self, b):
            """Up handler
            
            Args:
                b (ipywidgets.Button): clicked button
            """
            new_path = os.path.dirname(self.path.rstrip("/"))
            if new_path != self.path:
                self.path = new_path
            self.update()

        def on_click(self, change):
            """Click handler providing db and parent as context
            
            Args:
                db (str): database name
                parent (object): parent object
            """
            new_path = os.path.join(self.path, change["new"])
            if change["old"] is not None:
                if change["new"][-1] == "/":
                    self.path = new_path
                    self.update()
                else:
                    self.show_path(new_path)

        def close(self):
            """Close view"""
            self.sc.close()

    def __init__(self, dbutils=None):
        """Singleton initializer
        
        Args:
            dbutils (DBUtils, optional): DBUtils object. Defaults to None.
        """
        if not Dbfs.instance:
            Dbfs.instance = Dbfs.__Dbfs(dbutils)

    def __getattr__(self, name):
        """Singleton getattr overload"""
        return getattr(self.instance, name)


import configparser
import json
from jupyter_client import kernelspec
import os
import shutil
import subprocess
import sys
import textwrap
import time
from os.path import expanduser
from ssh_config import SSHConfig, Host
import inquirer

from databrickslabs_jupyterlab.utils import (bye, Dark, print_ok, print_error, print_warning)

# add all missing keys to ssh_config
Host.attrs += [
    ("Host", str),
    ("CanonicalizeFallbackLocal", str),
    ("CanonicalizeHostname", str),
    ("CanonicalizeMaxDots", str),
    ("CanonicalizePermittedCNAMEs", str),
    ("CASignatureAlgorithms", str),
    ("CertificateFile", str),
    ("ChallengeResponseAuthentication", str),
    ("CheckHostIP", str),
    ("Ciphers", str),
    ("ClearAllForwardings", str),
    ("Compression", str),
    ("ConnectionAttempts", str),
    ("ConnectTimeout", str),
    ("ControlMaster", str),
    ("ControlPath", str),
    ("ControlPersist", str),
    ("DynamicForward", str),
    ("EnableSSHKeysign", str),
    ("EscapeChar", str),
    ("ExitOnForwardFailure", str),
    ("FingerprintHash", str),
    ("ForwardX11", str),
    ("ForwardX11Timeout", str),
    ("ForwardX11Trusted", str),
    ("GatewayPorts", str),
    ("GlobalKnownHostsFile", str),
    ("GSSAPIAuthentication", str),
    ("GSSAPIDelegateCredentials", str),
    ("HashKnownHosts", str),
    ("HostbasedAuthentication", str),
    ("HostbasedKeyTypes", str),
    ("HostKeyAlgorithms", str),
    ("HostKeyAlias", str),
    ("IdentitiesOnly", str),
    ("IgnoreUnknown", str),
    ("Include", str),
    ("IPQoS", str),
    ("KbdInteractiveAuthentication", str),
    ("KbdInteractiveDevices", str),
    ("KexAlgorithms", str),
    ("MACs", str),
    ("NoHostAuthenticationForLocalhost", str),
    ("NumberOfPasswordPrompts", str),
    ("PasswordAuthentication", str),
    ("PermitLocalCommand", str),
    ("PKCS11Provider", str),
    ("ProxyJump", str),
    ("ProxyUseFdpass", str),
    ("PubkeyAcceptedKeyTypes", str),
    ("PubkeyAuthentication", str),
    ("RekeyLimit", str),
    ("RemoteCommand", str),
    ("RemoteForward", str),
    ("RequestTTY", str),
    ("RevokedHostKeys", str),
    ("SendEnv", str),
    ("ServerAliveCountMax", str),
    ("SetEnv", str),
    ("StreamLocalBindMask", str),
    ("StreamLocalBindUnlink", str),
    ("StrictHostKeyChecking", str),
    ("SyslogFacility", str),
    ("TCPKeepAlive", str),
    ("Tunnel", str),
    ("TunnelDevice", str),
    ("UpdateHostKeys", str),
    ("UseKeychain", str),
    ("UserKnownHostsFile", str),
    ("VerifyHostKeyDNS", str),
    ("VisualHostKey", str),
    ("XAuthLocation", str)
]

def utf8_decode(text):
    if isinstance(text, str):
        return text
        
    try:
        return text.decode("utf-8")
    except:
        # ok, let's replace the "bad" characters
        return text.decode("utf-8", "replace")

def execute(cmd):
    """Execute suprpcess
    
    Args:
        cmd (list(str)): Command as list of cmd parts (e.g. ["ls", "-l"])
    """
    try:
        # Cannot use encoding arg at the moment, since need to support python 3.5
        result = subprocess.run(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE).__dict__
        result["stderr"] = utf8_decode(result["stderr"])
        result["stdout"] = utf8_decode(result["stdout"])
        return result
    except Exception as ex:
        return {'args': cmd, 'returncode': -1, 'stdout': '', 'stderr': str(ex)}


def conda_version():
    """Check conda version"""
    result = execute(["conda", "--version"])
    if result["returncode"] != 0:
        print_error(result["stderr"])
        sys.exit(1)

    result = result["stdout"].strip()
    return tuple([int(v) for v in result.split(" ")[1].split(".")])


def write_config():
    """Store jupyter lab configuration necessary for databrickslabs_jupyterlab
    Set values:
    - c.KernelManager.autorestart: False
    - c.MappingKernelManager.kernel_info_timeout: 600
    """
    config = {"c.KernelManager.autorestart": False, "c.MappingKernelManager.kernel_info_timeout": 600}

    full_path = os.path.expanduser("~/.jupyter")
    if not os.path.exists(full_path):
        os.mkdir(full_path)

    config_file = os.path.join(full_path, "jupyter_notebook_config.py")
    if os.path.exists(config_file):
        with open(config_file, "r") as fd:
            lines = fd.read().split("\n")

        with open(config_file, "w") as fd:
            for line in lines:
                kv = line.strip().split("=")
                if len(kv) == 2:
                    k, v = kv
                    if config.get(k, None) is not None:
                        fd.write("%s=%s\n" % (k, config[k]))
                        del config[k]
                    else:
                        fd.write("%s\n" % line)
            for k, v in config.items():
                fd.write("%s=%s\n" % (k, v))
    else:
        with open(config_file, "w") as fd:
            fd.write("\n".join(["%s=%s" % (k, v) for k, v in config.items()]))


def get_local_libs():
    """Get installed libraries of the currtent conda environment
    """
    result = execute(["conda", "list", "--json"])
    if result["returncode"] != 0:
        print_error(result["stderr"])
        return False

    try:
        return json.loads(result["stdout"])
    except Exception as ex:
        print_error(ex)
        return False


def get_db_config(profile):
    """Get Databricks configuration from ~/.databricks.cfg for given profile
    
    Args:
        profile (str): Databricks CLI profile string
    
    Returns:
        tuple: The tuple of host and personal access token from ~/.databrickscfg
    """
    config = configparser.ConfigParser()
    configs = config.read(expanduser("~/.databrickscfg"))
    if not configs:
        print_error("Cannot read ~/.databrickscfg")
        bye(1)

    profiles = config.sections()
    if not profile in profiles:
        print(" The profile '%s' is not available in ~/.databrickscfg:" % profile)
        for p in profiles:
            print("- %s" % p)
        bye()
    else:
        host = config[profile]["host"]
        token = config[profile]["token"]
        return host, token

def add_known_host(public_dns, known_hosts="~/.ssh/known_hosts"):
    result = execute(["ssh-keygen", "-R", "[%s]:2200" % public_dns])
    result = execute(["ssh-keyscan", "-p", "2200", public_dns])
    if result["returncode"] == 0:
        fingerprint = result["stdout"]
        known_hosts_file = os.path.expanduser(known_hosts)
        with open(known_hosts_file, "a") as fd:
            fd.write("\n%s" % fingerprint)
        print("   => Known hosts fingerprint added for %s\n" % public_dns)
    else:
        print_warning("   => Could not add know_hosts fingerprint for %s\n" % public_dns)

def prepare_ssh_config(cluster_id, profile, public_dns):
    """Add/edit the ssh configuration belonging to the given cluster in ~/.ssh/config
    
    Args:
        cluster_id (str): Cluster ID
        profile (str): Databricks CLI profile string
        public_dns (str): Public DNS/IP address
    """
    backup_path = "~/.databrickslabs_jupyterlab/ssh_config_backup"

    config = os.path.expanduser("~/.ssh/config")
    if not os.path.exists(os.path.expanduser(backup_path)):
        os.makedirs(os.path.expanduser(backup_path))

    print_warning("   => ~/.ssh/config will be changed")
    backup = "%s/config.%s" % (backup_path, time.strftime("%Y-%m-%d_%H-%M-%S"))
    print_warning("   => A backup of the current ~/.ssh/config has been created")
    print_warning("   => at %s" % backup)

    shutil.copy(config, os.path.expanduser(backup))
    try:
        sc = SSHConfig.load(config)
    except:
        sc = SSHConfig(config)
    hosts = [h.name for h in sc.hosts()]
    if cluster_id in hosts:
        host = sc.get(cluster_id)
        host.set("HostName", public_dns)
        host.set("ServerAliveInterval", 5)
        host.set("ServerAliveCountMax", 2)
        host.set("ConnectTimeout", 5)
        print("   => Added ssh config entry or modified IP address:\n")
        print(textwrap.indent(str(host), "      "))
    else:
        # ServerAliveInterval * ServerAliveCountMax = 48h
        attrs = {
            "HostName": public_dns,
            "IdentityFile": "~/.ssh/id_%s" % profile,
            "Port": 2200,
            "User": "ubuntu",
            "ServerAliveInterval": 30,
            "ServerAliveCountMax": 5760,
            "ConnectTimeout": 5
        }
        host = Host(name=cluster_id, attrs=attrs)
        print("   => Adding ssh config to ~/.ssh/config:\n")
        print(textwrap.indent(str(host), "      "))
        sc.append(host)
    sc.write()

    add_known_host(public_dns)

def show_profiles():
    """Show locally configured Databricks CLI profile"""
    template = "%-20s %-60s %s"
    print("")
    print(template % ("PROFILE", "HOST", "SSH KEY"))
    config = configparser.ConfigParser()
    config.read(expanduser("~/.databrickscfg"))
    profiles = config.sections()

    for profile in profiles:
        host, _ = get_db_config(profile)
        ssh_ok = "OK" if os.path.exists(os.path.expanduser("~/.ssh/id_%s") % profile) else "MISSING"
        print(template % (profile, host, ssh_ok))


def create_kernelspec(profile, organisation, host, cluster_id, cluster_name, local_env, python_path):
    """Create or edit the ssh_ipykernel specification for jupyter lab
    
    Args:
        profile (str): Databricks CLI profile string    
        organisation (str): In case of Azure, the organization ID, else None
        host (str): host from databricks cli config for given profile string
        cluster_id (str): Cluster ID
        cluster_name (str): Cluster name
        local_env (str): Name of the local conda environment
        python_path (str): Remote python path to be used for kernel
    """
    from ssh_ipykernel.manage import add_kernel

    print("   => Creating kernel specification for profile '%s'" % profile)
    env = "DBJL_PROFILE=%s DBJL_HOST=%s DBJL_CLUSTER=%s" % (profile, host, cluster_id)
    if organisation is not None:
        env += " DBJL_ORG=%s" % organisation
    
    if cluster_name.replace(" ", "_") == local_env:
        display_name = "SSH %s %s:%s" % (cluster_id, profile, cluster_name)
    else:
        display_name = "SSH %s %s:%s (%s)" % (cluster_id, profile, cluster_name, local_env)

    add_kernel(
        host=cluster_id,
        display_name=display_name,
        local_python_path=sys.executable,
        remote_python_path=os.path.dirname(python_path),
        sudo=True,
        env=env,
        timeout=5
    )   
    print("   => Kernel specification 'SSH %s %s' created or updated" % (cluster_id, display_name))

def remove_kernelspecs():
    km = kernelspec.KernelSpecManager()

    kernel_id = None
    while kernel_id != "done":

        remote_kernels = {
            kernelspec.get_kernel_spec(k).display_name : k
            for k, v in kernelspec.find_kernel_specs().items() if k.startswith("ssh_")
        }
        if remote_kernels == {}:
            print_ok("   => No databricklabs_jupyterlab kernel spec left")
            break

        choice = [
            inquirer.List("kernel_name",
                        message="Which kernel spec to delete (Ctrl-C to finish)?",
                        choices=list(remote_kernels.keys()))
        ]
        answer = inquirer.prompt(choice, theme=Dark())

        if answer is None:
            break

        kernel_name = answer["kernel_name"]
        if kernel_id != "done":
            answer = input("Really delete kernels spec '%s' (y/n) " % kernel_name)
            if answer.lower() == "y":
                km.remove_kernel_spec(remote_kernels[kernel_name])



from collections import defaultdict
import json
import logging
import os
import socket
import sys
import threading
import time
from notebook.base.handlers import IPythonHandler
from tornado import web
from collections import defaultdict
from tornado.log import LogFormatter

import databrickslabs_jupyterlab
from databrickslabs_jupyterlab.remote import (connect, is_reachable, get_cluster, get_python_path, check_installed,
                                              install_libs)
from databrickslabs_jupyterlab.utils import SshConfig
from databrickslabs_jupyterlab.local import (get_db_config, prepare_ssh_config)
from databricks_cli.clusters.api import ClusterApi

if os.environ.get("DEBUG", None) is None:
    DEBUG_LEVEL = "INFO"
else:
    DEBUG_LEVEL = os.environ.get["DEBUG"]

_LOG_FMT = ("%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d " "%(name)s]%(end_color)s %(message)s")
_LOG_DATEFMT = "%H:%M:%S"

_console = logging.StreamHandler()
_console.setFormatter(LogFormatter(fmt=_LOG_FMT, datefmt=_LOG_DATEFMT))
_console.setLevel(DEBUG_LEVEL)

_logger = logging.getLogger("databrickslabs-jl")
_logger.propagate = False
_logger.setLevel(DEBUG_LEVEL)
_logger.handlers = []
_logger.addHandler(_console)


def get_cluster_state(profile, cluster_id):
    """Get cluster state
    
    Args:
        profile (str): Databricks CLI profile string
        cluster_id (str): Cluster ID
    
    Returns:
        dict: ClusterStatus
    """
    apiclient = connect(profile)
    client = ClusterApi(apiclient)
    cluster = client.get_cluster(cluster_id)
    state = cluster.get("state", None)
    # if state == "TERMINATED":
    #     _logger.warning("DbStatusHandler cluster state = %s" % state)
    # else:
    #     _logger.info("DbStatusHandler cluster state = %s" % state)

    return state


class Status:
    """Status implementation"""
    def __init__(self):
        self.status = defaultdict(lambda: defaultdict(lambda: {}))
        self.dots = 0
        self._installing = defaultdict(lambda: None)
        self._ignore_next_unreachable = False

    def get_status(self, profile, cluster_id):
        """Get current cluster start status for the jupyterlab status line
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
        
        Returns:
            str: Cluster status or "UNKNOWN"
        """
        if self.status[profile][cluster_id] == {}:
            # For the first run start_status is None.
            # Use REST API to provide a reasonable status
            state = get_cluster_state(profile, cluster_id)  # called max once
            if state in ["PENDING", "RESTARTING"]:
                status = "Starting"  # special status, don't set in global state
            elif state in ["RUNNING", "RESIZING"]:
                status = "Running"  # standard status
            elif state in ["TERMINATING", "TERMINATED"]:
                status = "UNREACHABLE"  # standard status
            else:
                status = state  # ERROR, UNKNOWN non standard status
            self.set_status(profile, cluster_id, status)

        return self.status[profile][cluster_id]

    def reset_status(self, profile, cluster_id):
        self.status[profile][cluster_id] = {}

    def set_status(self, profile, cluster_id, status, new_status=True):
        """Set Cluster start status for the jupyterlab status line
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
            status (str): Status value
            new_status (bool, optional): If True, number of progress dots is set to 0. Defaults to True.
        """
        self.dots = 0 if new_status else (self.dots + 1) % 6
        self.status[profile][cluster_id] = status + ("." * self.dots)

    def installing(self, profile, cluster_id):
        """Check whether cluster is currently in installation mode
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
        
        Returns:
            bool: True if in installation mode, else False
        """
        return self._installing["%s %s" % (profile, cluster_id)]

    def set_installing(self, profile, cluster_id):
        """Set installation mode for a cluster
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
        """
        self._installing["%s %s" % (profile, cluster_id)] = True

    def unset_installing(self, profile, cluster_id):
        """Unset installation mode for a cluster
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
        """
        self._installing["%s %s" % (profile, cluster_id)] = False
        self._ignore_next_unreachable = True

    def ignore_next_unreachable(self):
        result = self._ignore_next_unreachable
        self._ignore_next_unreachable = False
        return result


class KernelHandler(IPythonHandler):
    """Kernel handler to get jupyter kernel for given kernelspec"""
    nbapp = None
    status = Status()

    def get_kernel(self, kernel_id):
        """Get jupyter kernel for given kernel Id
        
        Args:
            kernel_id (str): Internal jupyter kernel ID
        
        Returns:
            KernelManager: KernelManager object
        """
        km = KernelHandler.nbapp.kernel_manager

        kernel_info = None
        for kernel_info in km.list_kernels():
            if kernel_info["id"] == kernel_id:
                break

        if kernel_info is not None:
            return km.get_kernel(kernel_info["id"])
        else:
            return None


class DbStatusHandler(KernelHandler):
    """Databricks cluster status handler"""
    @web.authenticated
    def get(self):
        """GET handler to return the current Databricks cluster start status"""
        global_status = KernelHandler.status

        profile = self.get_argument('profile', None, True)
        cluster_id = self.get_argument('cluster_id', None, True)
        kernel_id = self.get_argument('id', None, True)

        kernel = self.get_kernel(kernel_id)
        conn_info = {
            "control_port": kernel.control_port,
            "hb_port": kernel.hb_port,
            "iopub_port": kernel.iopub_port,
            "shell_port": kernel.shell_port,
            "stdin_port": kernel.stdin_port
        }
        
        from ssh_ipykernel.status import Status as KernelStatus
        kernel_status = KernelStatus(conn_info)

        status = None
        start_status = global_status.get_status(profile, cluster_id)

        if global_status.installing(profile, cluster_id):
            # While the DbStartHandler is installing the kernel, provide the global starting status
            status = start_status
        else:

            if kernel_status.is_running():
                status = "Connected"
            elif kernel_status.is_starting():
                status = "Starting"
            elif kernel_status.is_unknown():
                # Use REST API to determine status
                global_status.reset_status(profile, cluster_id)
                status = global_status.get_status(profile, cluster_id)
            elif kernel_status.is_connect_failed():
                status = "CONNECT FAILED"
            else:
                # just swallow the intermediate UNREACHABLE after a restart
                if global_status.ignore_next_unreachable():
                    status = "Starting"
                else:
                    status = "UNREACHABLE"

        result = {'status': "%s" % status}
        _logger.debug(
            "DbStatusHandler: installing: '%s'; start_status: '%s'; kernel_status: '%s'; status: '%s'" %
            (global_status.installing(profile, cluster_id), start_status, kernel_status.get_status_message(), result))
        self.finish(json.dumps(result))


class DbStartHandler(KernelHandler):
    """Databricks cluster start handler"""
    @web.authenticated
    def get(self):
        """GET handler to trigger cluster start in a separate thread"""
        profile = self.get_argument('profile', None, True)
        cluster_id = self.get_argument('cluster_id', None, True)
        kernel_id = self.get_argument('id', None, True)
        thread = threading.Thread(target=self.start_cluster, args=(profile, cluster_id, kernel_id))
        thread.daemon = True
        thread.start()

        result = {'status': "ok"}
        self.finish(json.dumps(result))

    def start_cluster(self, profile, cluster_id, kernel_id):
        """Start cluster in a separate thread
        
        Args:
            profile (str): Databricks CLI profile string
            cluster_id (str): Cluster ID
            kernel_id (str): Internal jupyter kernel ID
        """
        global_status = KernelHandler.status

        if global_status.installing(profile, cluster_id):
            _logger.info("DbStartHandler cluster %s:%s already starting" % (profile, cluster_id))
        else:
            _logger.info("DbStartHandler cluster %s:%s start triggered" % (profile, cluster_id))
            global_status.set_installing(profile, cluster_id)

            host, token = get_db_config(profile)
            cluster_id, public_ip, cluster_name, dummy = get_cluster(profile, host, token, cluster_id, global_status)
            if cluster_name is None:
                global_status.set_status(profile, cluster_id, "ERROR: Cluster could not be found")
                return

            global_status.set_status(profile, cluster_id, "Configuring SSH")
            prepare_ssh_config(cluster_id, profile, public_ip)
            if not is_reachable(public_dns=public_ip):
                global_status.set_status(profile, cluster_id, "UNREACHABLE")
            else:
                global_status.set_status(profile, cluster_id, "Installing driver libs")
                result = install_libs(cluster_id, host, token)
                if result[0] == 0:
                    _logger.info("DbStartHandler: installations done")
                else:
                    _logger.error("DbStartHandler: installations failed")
                    global_status.set_status(profile, cluster_id, "ERROR")

                time.sleep(1)
                kernel = self.get_kernel(kernel_id)
                kernel.restart_kernel(now=True)
                global_status.set_status(profile, cluster_id, "Running")
            global_status.unset_installing(profile, cluster_id)


def _jupyter_server_extension_paths():
    """
    Set up the server extension for status handling
    """
    return [{
        'module': 'databrickslabs_jupyterlab',
    }]
